# coding: utf-8
# 2015-01-14
# 1. 3ヶ月を超えたチェックはend_dateで行われていることを確認(エイムラック)
# 2. UIDは、AIPOのスケジュールのシーケンスなことを確認(エイムラック)
# 3. 繰り返しスケジュールは3ヶ月を無視して送ってくるが、AIPO側で変更予定があることを確認(エイムラック)
# 4. 10日ほど運用したが、open-uriでたまに接続エラーが発生するので、一応トラップしておく(サーバーもしょぼいし仕方ないかなぁ・・・
# 5. GoogleAPIのimportでレスポンス502(bad gateway)が発生したため、そのデータが削除に引っかかってしまった。エラーデータを削除対象にしない。
# 2015-01-19
# 1. 無期限の繰り返しスケジュールに設定されている3ヶ月後のUNTILをはずす
# 2. RRULEの組み立てロジック変更
# 2015-01-20
# 1. 終日スケジュールの終了日がAIPOの入力から+1日されている為、削除処理の3ヶ月判定が1日くるってしまったので、判定の日付は-1日する
# 2. テスト用にCalendarIdとTimeout値のローカル変数をちょっと変更
require 'date'
require 'icalendar'
require 'json'
require 'tzinfo'
require 'icalendar/recurrence'
require 'google/api_client'
require 'yaml'
require 'open-uri'
require 'resolv-replace'
require 'timeout'
require 'openssl'
module OpenSSL
  module SSL
    remove_const :VERIFY_PEER
  end
end

def event_start_date(event)
  if event.start.date
    event.start.date
  else
    event.start.dateTime.to_s.match(/^([\d]{4}-[\d]{2}-[\d]{2})\s([\d]{2}:[\d]{2})/)
    $1
  end
end
def event_end_date(event)
  if event.end.date
    event.end.date
  else
    event.end.dateTime.to_s.match(/^([\d]{4}-[\d]{2}-[\d]{2})\s([\d]{2}:[\d]{2})/)
    $1
  end
end

OpenSSL::SSL::VERIFY_PEER = OpenSSL::SSL::VERIFY_NONE


api_key = YAML.load_file('api.yaml')

TIME_OUT = api_key['timeout_time']
CALENDAR_ID = api_key['cal_id']
#AIPOからICALを取得。
#接続できなければ、後続処理はせず終了
uri = api_key['ical_uri']
certs = [api_key['aipo_user'], api_key['aipo_pass']]
ics = ""
begin
  timeout(TIME_OUT) do
    f = open(uri, {:http_basic_authentication => certs})
    ##icalendarがVTIMEZONEがあるとエラーを出すので、よくわからないからハズス
    writefl = 1
    f.each do |line|
      if line[0,15] == "BEGIN:VTIMEZONE"
        writefl = 0
      end
      if writefl == 1
        ics = ics + line
      end
      if line[0,13] == "END:VTIMEZONE"
        writefl = 1
      end  
    end      
  end
rescue TimeoutError => e
  printf("%s,%s\n","open-uri timeout err=",e)
  exit(1)  
rescue => e
  printf("%s,%s\n","open-uri error=",e)
  exit(1)
end

client = Google::APIClient.new(:application_name => '')
## Google Developers Console でサービス アカウントを取得
## Googleカレンダー側で、このサービスアカウントのメールアドレスを、アクセスするカレンダーに対して
## 事前に登録しておく
key = Google::APIClient::KeyUtils.load_from_pkcs12(api_key['key_file'], api_key['key_pass'])
client.authorization = Signet::OAuth2::Client.new(
  :token_credential_uri => 'https://accounts.google.com/o/oauth2/token',
  :audience => 'https://accounts.google.com/o/oauth2/token',
  :scope => 'https://www.googleapis.com/auth/calendar',
  :issuer => api_key['issuer'],
  :signing_key => key)

client.authorization.fetch_access_token!

service = client.discovered_api('calendar', 'v3')

ical = Icalendar::parse(ics,true)
#import errorが起こったスケジュールが消されてしまうので、
#uidをプールする仕組みを追加して、削除処理で判定する
erruids = []
  
# icsを加工しながら連想配列化する
ical.events.each do |icalevent|  
  allday = 0    
  # 検索開始・終了日時の設定
  # 終日スケジュールはDate型で入っている
  if icalevent.dtstart.kind_of?(Icalendar::Values::Date)
   # 終日の予定の場合
    allday = 1
    str_date_start = icalevent.dtstart.strftime("%Y-%m-%d")
    str_date_end = icalevent.dtend.strftime("%Y-%m-%d")
  else
    # 時刻指定の予定の場合 （末尾にtimezoneを指定しないと、うまく検索されない）
    str_date_start = icalevent.dtstart.strftime("%Y-%m-%dT%H:%M:%S.000+09:00")
    str_date_end = icalevent.dtend.strftime("%Y-%m-%dT%H:%M:%S.000+09:00")        
  end

  # 繰り返しスケジュールはrruleの有無で判定
  # AIPOは幸い曜日繰り返しくらいしかないので、RRULEなしか、RRULE1つしかないと決め付け
  str_rrule = ""
  str_exdate = ""
  isrecurrence = 0  
  if icalevent.rrule == []
    isrecurrence = 0
  else
    isrecurrence = 1
    #期限のない繰り返しスケジュールの期限(UNTIL)をAIPO側で3ヶ月後に設定されている
    #データ作成日の3ヶ月後と期限(UNTIL)が一致したら、RRULEのUNTILを削除する
    #※これの弊害で、期限が設定されている場合も1日だけ無期限になる日が発生する
    #ついでに苦手な正規表現化・・・
    stamptime = Time.parse(icalevent.dtstamp.to_s)
    #UTCで入っているので、ローカルタイムで補正
    stampdate = Date.strptime(stamptime.localtime.to_s)
    limitdate = stampdate >> 3
    icalevent.rrule.each do |ru|
      rrules = ru.value_ical.split(/\s*;\s*/)      
      untildate = Date.strptime(ru.until,"%Y%m%dT%H%M%S")
      rrules.each do |rule|
        if rule.match(/^UNTIL/)
          if limitdate == untildate
            rule = ""
          else
            # UNITLの日時指定の末尾にZ(UTC)を付けないとエラーだったので追加
            rule = rule + "Z" 
          end                   
        end
        
        if str_rrule != "" && rule != ""
          str_rrule = str_rrule + ";" 
        end
        str_rrule = str_rrule + rule
      end
      str_rrule = "RRULE:" + str_rrule        
    end             

    #除外日はexdateの有無で判定。加工なしで利用できた
    isexdate = 0
    if icalevent.exdate == []
      isexdate = 0
    else
      isexdate = 1
      icalevent.exdate.each do |ex|
        str_exdate = "EXDATE:" + ex.value_ical
      end
    end        
  end   
  # UIDの加工
  # AIPOのアクセス日が追加されているのではずす
  # UIDが変わると、スケジュールが追加されていってしまう。
  # ユニーク性は調査していないが、きっとISNやシーケンスなんだろうと決めつけ利用
  # ↑AIPO上のscheduleIdから作っていることを確認。シーケンスですね。
  str_uid_work = icalevent.uid.to_s
  len = str_uid_work.size
  pos = str_uid_work.index("-")
  str_uid = str_uid_work[pos+1,len-pos]

  event = {}
  event['summary'] = icalevent.summary
  event['location'] = icalevent.location
  event['description'] = icalevent.description
  event['iCalUID'] = str_uid
  if allday == 0
    h_sta = {
      'dateTime' => str_date_start,
      'timeZone' => api_key['timezone']
    }
    h_end = {
      'dateTime' => str_date_end,
      'timeZone' => api_key['timezone']
    }
  else
    h_sta = {
      'date' => str_date_start,
      'timeZone' => api_key['timezone']
    }
    h_end = {
      'date' => str_date_end,
      'timeZone' => api_key['timezone']
    }   
  end  
  event['start'] = h_sta
  event['end'] = h_end
  if isrecurrence == 1
    if isexdate == 0
      event['recurrence'] = [ str_rrule ]
    else
      event['recurrence'] = [ str_rrule,str_exdate ]
    end
  end
## importとinsertの違いはよくわからないが、uidが同じなら更新してくれるようなのでimportを使用
  result = client.execute(:api_method => service.events.import,
                          :parameters => {'calendarId' => CALENDAR_ID},
                          :body => JSON.dump(event),
                          :headers => {'Content-Type' => 'application/json'})
 
  if result.status != 200
    erruids.push(str_uid)
    printf("%s,%s,%s\n","GAPI import err status=",result.status,JSON.dump(event))
  end
end
# 更新が終わったところで、AIPO側で削除されたデータを削除する
# 更新前に全件削除も考えたが、deleteに2〜3分、importに2〜3分かかるため
# たまたまGoogleカレンダーをみたユーザーが勘違いしそうなケースが増えそうなので
# 極力、削除されたデータだけを削除する
events = []
page_token = nil
# Googleカレンダーのデータを全件読み込み
begin
  params = {}
  params = {'calendarId' => CALENDAR_ID}
  if page_token.to_s != ''
      params['pageToken'] = page_token
  end
  result = client.execute(:api_method => service.events.list,
                            :parameters => params)
  if result.status == 200
    result.data.items.each do |item|
      events << item
    end      
    page_token = result.data.next_page_token
  else
    page_token = nil
  end                              
end while page_token.to_s != ''

# 削除判定
# 基本削除ルールは、上記Importで更新されなかったデータは削除
# CRONを10分毎にする予定なので、10分以上updatedが古いデータ
# ただし、AIPOは3ヶ月を超えたデータは送ってこないので、削除判定不能のため無視。(過去なので、まぁーいいでしょ)
# 繰り返しスケジュールは、3ヶ月を超えた古いデータも送ってきている模様
# この3ヶ月はもしかしたらend_dateなのかも知れないので、あとで要チェック(過去なので、まぁーいいでしょ)
# ↑end_dateでの判定だそうで変更する
nowdate = Date.today 
events.each do |event|
  ret = catch(:del_check) {
    #importでエラーが起こったuidは削除しない
    if erruids.include?(event.iCalUID) == true
      throw :del_check,nil
    end
    
    mins = (Time.now.gmtime - event.updated).divmod(60)
    #終日スケジュールは、AIPO上の日付に+1されていることが判明
    #終日の場合は、-1日する
    limitdate = Date.strptime(event_end_date(event), "%Y-%m-%d") >> 3 # 3ヶ月加算
    if event.start.date
      
      limitdate = limitdate - 1 #1日減算
      ## printf("%s,%s,%s\n",event_start_date(event),event_end_date(event),limitdate)
    end
        
    if mins[0] > 10
      
      if event.recurrence == []
        #3ヶ月を経過するとAIPOは送ってこなくなるので、判別不能
        #正確には、時間も考慮しなければいけないのかも知れないけど、過去なので、1日くらいまぁーいいでしょ
        if nowdate < limitdate
          throw :del_check,1
        end        
      else
        #繰り返しは古くても送られてくるようなので、更新が古ければ削除
        throw :del_check,1
      end
    end  
  }  

  if ret == 1
    params = {}
    params = {'calendarId' => CALENDAR_ID,
              'eventId' => event.id
    }
    result = client.execute(:api_method => service.events.delete,
                            :parameters => params)
   
    #削除時は204が正常の様子。念の為、200も残しておく         
    if !(result.status == 200)
      if (result.status == 204)
        printf("%s,%s,%s,%s\n","GAPI delete info status=",result.status,event_start_date(event),event.summary)
      else
        printf("%s,%s,%s,%s\n","GAPI delete err status=",result.status,event_start_date(event),event.summary)
      end             
    end
  end
end
